<?php

namespace Drupal\test_helpers;

use Drupal\Component\Annotation\Doctrine\SimpleAnnotationReader;
use Drupal\Core\Cache\CacheTagsInvalidatorInterface;
use Drupal\Core\Database\Query\ConditionInterface as QueryConditionInterface;
use Drupal\Core\DependencyInjection\ContainerBuilder;
use Drupal\Core\Entity\EntityInterface;
use Drupal\Core\Entity\EntityStorageInterface;
use Drupal\Core\Entity\Query\ConditionInterface as EntityQueryConditionInterface;
use Drupal\test_helpers\Stub\ModuleHandlerStub;
use Drupal\test_helpers\Stub\TokenStub;
use PHPUnit\Framework\Assert;
use PHPUnit\Framework\MockObject\MockObject;
use Symfony\Component\DependencyInjection\Exception\ServiceNotFoundException;
use Symfony\Component\Yaml\Yaml;

/**
 * Helper functions to simplify writing of Unit Tests.
 */
class UnitTestHelpers {

  /**
   * The list of implemented custom stubs for services.
   */
  const SERVICES_CUSTOM_STUBS = [
    'entity_type.manager' => EntityTypeManagerStub::class,
    'database' => DatabaseStub::class,
    'token' => TokenStub::class,
    'module_handler' => ModuleHandlerStub::class,
  ];

  /**
   * Gets a protected method from a class using reflection.
   */
  public static function getProtectedMethod(object $class, string $methodName): \ReflectionMethod {
    $reflection = new \ReflectionClass($class);
    $method = $reflection
      ->getMethod($methodName);
    $method
      ->setAccessible(TRUE);
    return $method;
  }

  /**
   * Calls a protected method from a class using reflection.
   */
  public static function callProtectedMethod(object $class, string $methodName, array $arguments = []) {
    $method = self::getProtectedMethod($class, $methodName);
    return $method->invokeArgs($class, $arguments);
  }

  /**
   * Gets a protected property from a class using reflection.
   */
  public static function getProtectedProperty(object $class, string $propertyName) {
    $reflection = new \ReflectionClass($class);
    $property = $reflection
      ->getProperty($propertyName);
    $property
      ->setAccessible(TRUE);
    return $property->getValue($class);
  }

  /**
   * Parses the annotation for a class and gets the definition.
   */
  public static function getPluginDefinition(string $class, string $plugin, string $annotationName = NULL) {
    static $definitions;

    if (isset($definitions[$plugin][$class])) {
      return $definitions[$plugin][$class];
    }

    $rc = new \ReflectionClass($class);

    $reader = new SimpleAnnotationReader();
    $reader->addNamespace('Drupal\Core\Annotation');
    $reader->addNamespace('Drupal\Core\\' . $plugin . '\Annotation');

    // If no annotation name is passed, just getting the first anotation.
    if (!$annotationName) {
      $annotation = current($reader->getClassAnnotations($rc));
    }
    else {
      $annotation = $reader->getClassAnnotation($rc, $annotationName);
    }
    if ($annotation) {
      // Inline copy of the proteced function
      // AnnotatedClassDiscovery::prepareAnnotationDefinition().
      $annotation->setClass($class);

      $definition = $annotation->get();

      return $definition;
    }
  }

  /**
   * Binds a closure function to a mocked class method.
   */
  public static function bindClosureToClassMethod(\Closure $closure, MockObject $class, string $method): void {
    $doClosure = $closure->bindTo($class, get_class($class));
    $class->method($method)->willReturnCallback($doClosure);
  }

  /**
   * Tests simple create() and __construct() functions.
   *
   * @param string|object $class
   *   The class to test, can be a string with path or initialized class.
   * @param array $createArguments
   *   The list of arguments for passing to function create().
   *
   * @return object
   *   The initialized class instance.
   */
  public static function doTestCreateAndConstruct($class, array $createArguments = []): object {
    $container = UnitTestHelpers::getContainer();
    $classInstance = $class::create($container, ...$createArguments);
    $className = is_string($class) ? $class : get_class($class);
    Assert::assertInstanceOf($className, $classInstance);
    return $classInstance;
  }

  /**
   * Performs matching of passed conditions with the query.
   */
  public static function queryIsSubsetOf(object $query, object $queryExpected, $onlyListed = FALSE): bool {
    // @todo add checks for range, sort and other query parameters.
    return self::matchConditions($query->condition, $queryExpected->condition, $onlyListed);
  }

  /**
   * Performs matching of passed conditions with the query.
   */
  public static function matchConditions(object $conditionsObject, object $conditionsExpectedObject, $onlyListed = FALSE): bool {
    if ($conditionsObject instanceof EntityQueryConditionInterface) {
      if (strcasecmp($conditionsObject->getConjunction(), $conditionsExpectedObject->getConjunction()) != 0) {
        return FALSE;
      }
    }
    elseif ($conditionsObject instanceof QueryConditionInterface) {
      if (strcasecmp($conditionsObject->conditions()['#conjunction'], $conditionsExpectedObject->conditions()['#conjunction']) != 0) {
        return FALSE;
      }
    }
    else {
      throw new \Exception("Conditions should implement Drupal\Core\Entity\Query\ConditionInterface or Drupal\Core\Database\Query\ConditionInterface.");
    }
    $conditions = $conditionsObject->conditions();
    unset($conditions['#conjunction']);
    $conditionsExpected = $conditionsExpectedObject->conditions();
    unset($conditionsExpected['#conjunction']);
    $conditionsFound = [];
    foreach ($conditions as $condition) {
      foreach ($conditionsExpected as $conditionsExpectedDelta => $conditionExpected) {
        if (is_object($condition['field']) || is_object($conditionExpected['field'])) {
          if (!is_object($condition['field']) || !is_object($conditionExpected['field'])) {
            continue;
          }
          return self::matchConditions($condition['field'], $conditionExpected['field'], $onlyListed);
        }
        if (self::isNestedArraySubsetOf($condition, $conditionExpected)) {
          $conditionsFound[$conditionsExpectedDelta] = TRUE;
        }
      }
    }
    if (count($conditionsFound) != count($conditionsExpected)) {
      return FALSE;
    }
    if ($onlyListed && (count($conditions) != count($conditionsExpected))) {
      return FALSE;
    }
    return TRUE;
  }

  /**
   * Performs check if the actial array is a subset of expected.
   */
  public static function isNestedArraySubsetOf($array, $subset): bool {
    if (!is_array($array) || !is_array($subset)) {
      return FALSE;
    }
    $result = array_uintersect_assoc($subset, $array, self::class . '::isValueSubsetOfCallback');
    return $result == $subset;
  }

  /**
   * Gets a Drupal services container, or creates a new one.
   */
  public static function getContainer($forceCreate = FALSE): object {
    $container = (!$forceCreate && \Drupal::hasContainer())
      ? \Drupal::getContainer()
      : new ContainerBuilder();
    \Drupal::setContainer($container);
    return $container;
  }

  /**
   * Adds a new service to the Drupal container, if exists - reuse existing.
   */
  public static function addToContainer(string $serviceName, object $class, bool $override = FALSE): object {
    $container = self::getContainer();
    $currentService = $container->has($serviceName)
      ? $container->get($serviceName)
      : new \stdClass();
    if (
      (get_class($currentService) !== get_class($class))
      || $override
    ) {
      $container->set($serviceName, $class);
    }
    return $container->get($serviceName);
  }

  /**
   * Gets a service class by name, using Drupal defaults or a custom YAML file.
   */
  public static function getServiceClassByName(string $serviceName, string $servicesYamlFile = NULL): string {
    if ($servicesYamlFile) {
      // @phpcs-ignore
      $services = Yaml::parseFile(DRUPAL_ROOT . '/' . $servicesYamlFile)['services'];
      $serviceClass = $services[$serviceName]['class'] ?? FALSE;
    }
    else {
      require_once dirname(__FILE__) . '/includes/DrupalCoreServicesMap.data';
      // @php-ignore
      $serviceClass = DRUPAL_CORE_SERVICES_MAP[$serviceName] ?? FALSE;
    }
    if (!$serviceClass) {
      throw new \Exception("Service '$serviceName' is missing in the list.");
    }
    return $serviceClass;
  }

  /**
   * Creates a mock for a service with getting definition from a YAML file.
   */
  public static function createServiceMock(string $serviceName, string $servicesYamlFile = NULL): MockObject {
    $serviceClass = self::getServiceClassByName($serviceName, $servicesYamlFile);
    $service = UnitTestHelpers::createMock($serviceClass);
    self::addToContainer($serviceName, $service);
    return $service;
  }

  /**
   * Gets a service stub: custom stub or just a mock for a default Drupal class.
   */
  public static function getServiceStub(string $serviceName, bool $onlyCustomMocks = FALSE): object {
    $container = UnitTestHelpers::getContainer();
    if ($container->has($serviceName)) {
      return $container->get($serviceName);
    }
    $service = self::getServiceStubClass($serviceName, $onlyCustomMocks);
    $container->set($serviceName, $service);
    return $service;
  }

  /**
   * Gets the class for a service, including current module implementations.
   */
  public static function getServiceStubClass(string $serviceName, bool $onlyCustomMocks = FALSE): object {
    if (isset(self::SERVICES_CUSTOM_STUBS[$serviceName])) {
      $serviceClass = self::SERVICES_CUSTOM_STUBS[$serviceName];
      $service = new $serviceClass();
    }
    elseif ($onlyCustomMocks) {
      throw new ServiceNotFoundException($serviceName);
    }
    else {
      $service = UnitTestHelpers::createServiceMock($serviceName);
    }
    return $service;
  }

  /**
   * Gets the service from the Drupal container, or creates a new one.
   */
  public static function getOrCreateService(string $serviceName, $class): object {
    $container = UnitTestHelpers::getContainer();
    if (!$container->has($serviceName)) {
      $container->set($serviceName, $class);
      \Drupal::setContainer($container);
    }
    return $container->get($serviceName);
  }

  /**
   * Creates a stub for an entity from a given class.
   */
  public static function createEntityStub(string $entityClassName, array $values = [], array $options = []): EntityInterface {
    $entityStorage = self::getEntityStorageStub($entityClassName);
    return $entityStorage->stubCreateEntity($entityClassName, $values, $options);
  }

  /**
   * Gets or initializes an Entity Storage for a given Entity class name.
   */
  public static function getEntityStorageStub(string $entityClassName): EntityStorageInterface {
    return self::getServiceStub('entity_type.manager')->stubGetOrCreateStorage($entityClassName);
  }

  /**
   * Initializes the main services to get mocks for entities.
   */
  public static function initEntityTypeManagerStubs(): void {
    self::getServiceStub('entity_type.manager');
  }

  /* ************************************************************************ *
   * Wrappers for UnitTestCase functions to make them available statically.
   * ************************************************************************ */

  /**
   * Gets the random generator for the utility methods.
   */
  public static function getRandomGenerator() {
    return UnitTestCaseWrapper::getInstance()->getRandomGenerator();
  }

  /**
   * Sets up a container with a cache tags invalidator.
   */
  public static function getContainerWithCacheTagsInvalidator(CacheTagsInvalidatorInterface $cache_tags_validator) {
    return UnitTestCaseWrapper::getInstance()->getContainerWithCacheTagsInvalidator($cache_tags_validator);
  }

  /**
   * Returns a stub class resolver.
   */
  public static function getClassResolverStub() {
    return UnitTestCaseWrapper::getInstance()->getClassResolverStub();
  }

  /**
   * Returns a stub translation manager that just returns the passed string.
   */
  public static function getStringTranslationStub() {
    return UnitTestCaseWrapper::getInstance()->getStringTranslationStub();
  }

  /**
   * Returns a mock object for the specified class.
   */
  public static function createMock(string $originalClassName): MockObject {
    return UnitTestCaseWrapper::getInstance()->createMock($originalClassName);
  }

  /**
   * Returns a partial mock object for the specified class.
   */
  public static function createPartialMock(string $originalClassName, array $methods): MockObject {
    return UnitTestCaseWrapper::getInstance()->createPartialMock($originalClassName, $methods);
  }

  /* ************************************************************************ *
   * UnitTestCase additions.
   * ************************************************************************ */

  /**
   * Creates a partial mock for the class and call constructor with arguments.
   */
  public static function createPartialMockWithConstructor(string $originalClassName, array $methods, array $constructorArgs = [], array $addMethods = NULL): MockObject {
    return UnitTestCaseWrapper::getInstance()->createPartialMockWithConstructor($originalClassName, $methods, $constructorArgs, $addMethods);
  }

  /**
   * Creates a partial mock with ability to add custom methods.
   */
  public static function createPartialMockWithCustomMethods(string $originalClassName, array $methods, array $addMethods = NULL): MockObject {
    return UnitTestCaseWrapper::getInstance()->createPartialMockWithCustomMethods($originalClassName, $methods, $addMethods);
  }

  /* ************************************************************************ *
   * Internal functions.
   * ************************************************************************ */

  /**
   * Internal callback helper function for array_uintersect.
   */
  private static function isValueSubsetOfCallback($value, $expected): int {
    // The callback function for array_uintersect should return
    // integer instead of bool (-1, 0, 1).
    if (is_array($expected)) {
      return self::isNestedArraySubsetOf($value, $expected) ? 0 : -1;
    }
    return ($value == $expected) ? 0 : -1;
  }

  /**
   * Disables a construtor calls to allow only static calls.
   */
  private function __construct() {
  }

}
